﻿#include <iostream>
#include <thread>
#include <mutex>
#include "Graphics/Graphics.h"
#include "Utils/Utils.h"
#include "Media/H264Encoder.h"
#include "DXGIDuplicator.h"

#include "ImGui/imgui.h"
#include "imgui/imgui_impl_win32.h"
#include "imgui/imgui_impl_dx11.h"
#include "ImHelper.h"
#include "CursorData.h"

#pragma comment(lib, "d3d11.lib")
#pragma comment(lib, "d3dcompiler.lib")

#pragma comment(lib, "d3d11.lib")
#pragma comment(lib, "d3dcompiler.lib")
//#pragma comment( linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"" )

#define u8s(x) reinterpret_cast<const char*>(u8##x)


HWND hwnd = NULL;
ID3D11Device* pD3DDevice_Duplicator = nullptr;
ID3D11DeviceContext* pContext_Duplicator = nullptr;
ID3D11Device* g_pd3dDevice = NULL;
ID3D11DeviceContext* g_pImmediateContext = NULL;
ID3D11RenderTargetView* g_pRenderTargetView = NULL;
IDXGISwapChain* g_pSwapChain = NULL;
DXGIDuplicator* duplicator = NULL;
ID3D11Texture2D* desktopTexture2D = NULL;
IWICBitmap* WIC = NULL;
//鼠标样式表
std::vector <Tuple<HANDLE, const CursorDefine&>> CURSOR_DATA = std::vector<Tuple<HANDLE, const CursorDefine&>>();
#define GENERATE_CURSOR_DATA(d)     CURSOR_DATA.push_back(Tuple<HANDLE, const CursorDefine&>(LoadCursor(NULL,MAKEINTRESOURCE(d)), _CURSOR_##d))

LONGLONG StartTime = 0;

//是否正在录制
bool _CaptureFrames = false;
//是否录制音频
bool captureSound = true;
//
int FPS = 25;

std::wstring MP4_PATH = L"./output.mp4";
//系统音频输出捕获对象
AudioCapture* _audioCap = NULL;
//H264编码对象
H264Encoder* _encoder = NULL;
//音频数据缓存
BYTE _buffer_Audio[0x40000]{};
//空白音频
BYTE _zero_buffer_Audio[38400]{};

//码率(影响视频质量和文件大小)
int VIDEOBITRATE = 0;
const char* VIDEOBITRATE_Names[] =
{
    "4M",
    "8M",
    "16M",
    "32M",
    "64M",
    "128M",
    "256M"
};
const int VIDEOBITRATE_Values[] =
{
    4000000,
    8000000,
    16000000,
    32000000,
    64000000,
    128000000,
    256000000,
};

DWORD capture_AudioThread(LPVOID lpThreadParameter)
{
    AudioCapture* _audioCap = (AudioCapture*)lpThreadParameter;
    while (_CaptureFrames)
    {
        UINT32 bufferSize = 0;
        LONGLONG duration = 0;
        auto hr = _audioCap->Capture(_buffer_Audio, &bufferSize, &duration);
        if (SUCCEEDED(hr) && bufferSize > 0)
            _encoder->AddAudioFrame(_buffer_Audio, bufferSize, duration);
        else
            std::this_thread::sleep_for(std::chrono::microseconds(10));
    }
    return NULL;
}
DWORD captureThread(LPVOID lpThreadParameter)
{
    ID3D11Texture2D* cnv = NULL;
    int screenWidth = duplicator->Width;
    int screenHeight = duplicator->Height;

    WIC = Factory::CreateWICBitmap(duplicator->Width, duplicator->Height);
    _audioCap = new AudioCapture();

    //mp4编码对象
    _encoder = new H264Encoder(MP4_PATH, screenWidth, screenHeight, FPS, VIDEOBITRATE_Values[VIDEOBITRATE]);

    //设置流音频格式
    _encoder->SetCurrentAudioType(_audioCap->GetMediaType());

    //开始
    _encoder->Begin();
    _audioCap->Start();
    int currFrame = 0;
    auto startTime = std::chrono::high_resolution_clock::now();
    StartTime = GetTickCount64();
    CreateThread(
        NULL, NULL, capture_AudioThread, _audioCap, NULL, NULL
    );
    while (_CaptureFrames)
    {
        HRESULT hr = duplicator->GetDesktopFrame(desktopTexture2D);
        if (SUCCEEDED(hr)) {
            WICRect lockRect = { 0, 0, screenWidth, screenHeight };
            IWICBitmapLock* pLock = NULL;
            hr = WIC->Lock(&lockRect, WICBitmapLockWrite, &pLock);
            if (SUCCEEDED(hr))
            {
                UINT cbBufferSize = 0;
                BYTE* pPixels = NULL;
                hr = pLock->GetDataPointer(&cbBufferSize, &pPixels);
                if (SUCCEEDED(hr))
                {
                    //转换一下格式，否则Map失败(事实上是为了将GPU内存Map到了CPU内存中做准备转换后的对象具备了D3D11_USAGE_STAGING)
                    cnv = duplicator->ConvertFormat(desktopTexture2D, DXGI_FORMAT_B8G8R8A8_UNORM);
                    BYTE* buffer = (BYTE*)duplicator->MapBuffer(cnv);
                    int dst_rowpitch = screenWidth * 4;
                    for (int h = 0; h < screenHeight; h++) {
                        memcpy(pPixels + h * dst_rowpitch, buffer + ((screenHeight - (h + 1)) * dst_rowpitch), dst_rowpitch);
                    }
                    duplicator->UnMapBuffer(cnv);
#pragma region 绘制鼠标光标
                    uint32_t* pixels = (uint32_t*)pPixels;
                    auto _setpixel = [&](int xx, int yy, bool mdown)
                        {
                            yy = screenHeight - (yy + 1);
                            if (xx < screenWidth && yy < screenHeight)
                            {
                                uint32_t oldC = pixels[(yy * screenWidth) + xx];
                                if (mdown)
                                    pixels[(yy * screenWidth) + xx] = RGB(0, 0, 255);
                                else
                                    pixels[(yy * screenWidth) + xx] = RGB(255 - GetRValue(oldC), 255 - GetGValue(oldC), 255 - GetBValue(oldC));
                            }
                        };
                    auto drawMouseArrow = [&](int startX, int startY, bool mdown, const CursorDefine& cursorPattern)
                        {
                            if (cursorPattern.Center)
                            {
                                startX -= 16;
                                startY -= 16;
                            }
                            for (int y = 0; y < 32; ++y)
                            {
                                for (int x = 0; x < 32; ++x)
                                {
                                    if (cursorPattern.Data[y] & (1 << (31 - x)))
                                    {
                                        _setpixel(startX + x, startY + y, mdown);
                                    }
                                }
                            }
                        };
                    bool mdown = GetAsyncKeyState(VK_LBUTTON) & 0x8000;
                    CURSORINFO cursorInfo = { sizeof(CURSORINFO) };
                    GetCursorInfo(&cursorInfo);
                    HCURSOR hCursor = cursorInfo.hCursor;
                    for (auto cur : CURSOR_DATA)
                    {
                        if (hCursor == cur.Item1)
                        {
                            drawMouseArrow(cursorInfo.ptScreenPos.x, cursorInfo.ptScreenPos.y, mdown, cur.Item2);
                            hCursor = NULL;
                            break;
                        }
                    }
                    if (hCursor) drawMouseArrow(cursorInfo.ptScreenPos.x, cursorInfo.ptScreenPos.y, mdown, CURSOR_DATA[0].Item2);
#pragma endregion
                }
                pLock->Release();
            }

            _encoder->AddVideoFrame(WIC);
            desktopTexture2D->Release();

        }
        else if (cnv)
        {
            _encoder->AddVideoFrame(WIC);
        }
        currFrame++;
        while ((_encoder->VideoTimeStamp - _encoder->AudioTimeStamp) >= 100'000)
        {
            _encoder->AddAudioFrame(_zero_buffer_Audio, 3840, 100'000);
        }
        //等待FPS间隔
        while (1)
        {
            auto now = std::chrono::high_resolution_clock::now();
            auto duration = std::chrono::duration_cast<std::chrono::microseconds>(now - startTime).count();
            ULONGLONG frameTime = currFrame * (1000000 / FPS);
            if (duration >= frameTime)
                break;
            else
                std::this_thread::sleep_for(std::chrono::microseconds(10));
        }
    }
    //结束全部流，关闭音频捕获
    _encoder->EndAudio();
    _encoder->End();
    _audioCap->Stop();
    delete _encoder;
    delete _audioCap;
    WIC->Release();
    WIC = NULL;
    return NULL;
}
void RenderGui(SIZE windowSize)
{
    ImGui::Text(u8s("帧率:"));
    ImGui::SameLine();
    ImGui::SetNextItemWidth(windowSize.cx - (ImGui::GetCursorPos().x + 5));
    ImGui::InputInt("##input_FPS", &FPS, 1, 100, _CaptureFrames ? ImGuiInputTextFlags_ReadOnly : 0);

    ImGui::Text(u8s("码率:"));
    ImGui::SameLine();
    auto pos = ImGui::GetCursorPos();
    ImGui::SetNextItemWidth(windowSize.cx - (ImGui::GetCursorPos().x + 5));
    if (_CaptureFrames)
    {
        ImGui::Text(VIDEOBITRATE_Names[VIDEOBITRATE]);
    }
    else
    {

        if (ImGui::BeginCombo("##GenerateAudio_GPT_weights_Sel", VIDEOBITRATE_Names[VIDEOBITRATE])) {
            for (int n = 0; n < 7; n++) {
                bool is_selected = (VIDEOBITRATE == n);
                if (ImGui::Selectable(VIDEOBITRATE_Names[n], is_selected)) {
                    VIDEOBITRATE = n;
                }
                if (is_selected) {
                    ImGui::SetItemDefaultFocus();
                }
            }
            ImGui::EndCombo();
        }
    }

    ImGui::Checkbox(u8s("录制声音##CaptureSound"), &captureSound);
    ImGui::SameLine();

    if (!_CaptureFrames)
    {
        if (ImGui::Button(u8s("开始##CaptureStart"), ImVec2(windowSize.cx - (ImGui::GetCursorPos().x + 5), 0)))
        {
            _CaptureFrames = true;
            auto cTime = DateTime::Now().ToString();
            cTime = StringHelper::Replace(cTime, "-", "_");
            cTime = StringHelper::Replace(cTime, " ", "_");
            cTime = StringHelper::Replace(cTime, ".", "_");
            cTime = StringHelper::Replace(cTime, ":", "_");

            MP4_PATH = Convert::string_to_wstring(cTime + ".mp4");
            auto dskPath = Environment::GetFolderPath(SpecialFolder::Desktop);
            MP4_PATH = Convert::string_to_wstring(dskPath) + L"\\" + MP4_PATH;

            CreateThread(
                NULL, NULL, captureThread, NULL, NULL, NULL
                //NULL, NULL, captureWin32Thread, NULL, NULL, NULL
            );
            StartTime = GetTickCount64() * 2;
        }
    }
    else
    {
        if (ImGui::Button(u8s("停止##CaptureEnd"), ImVec2(windowSize.cx - (ImGui::GetCursorPos().x + 5), 0)))
        {
            _CaptureFrames = false;
        }
    }
    if (_CaptureFrames)
    {
        LONGLONG now = GetTickCount64();
        if (now > StartTime)
        {
            static ID3D11Texture2D* texture = NULL;
            static ID3D11ShaderResourceView* image = NULL;
            if (WIC)
            {
                if (texture == NULL)
                {
                    D3D11_TEXTURE2D_DESC texDesc = {};
                    ZeroMemory(&texDesc, sizeof(texDesc));
                    texDesc.Width = GetSystemMetrics(SM_CXSCREEN);
                    texDesc.Height = GetSystemMetrics(SM_CYSCREEN);
                    texDesc.MipLevels = 1;
                    texDesc.ArraySize = 1;
                    texDesc.Format = DXGI_FORMAT_B8G8R8A8_UNORM;
                    texDesc.SampleDesc.Count = 1;
                    texDesc.Usage = D3D11_USAGE_DYNAMIC;
                    texDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
                    texDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
                    texDesc.MiscFlags = 0;

                    g_pd3dDevice->CreateTexture2D(&texDesc, nullptr, &texture);

                    // 创建着色器资源视图
                    D3D11_SHADER_RESOURCE_VIEW_DESC srvDesc;
                    ZeroMemory(&srvDesc, sizeof(srvDesc));
                    srvDesc.Format = texDesc.Format;
                    srvDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
                    srvDesc.Texture2D.MipLevels = texDesc.MipLevels;
                    srvDesc.Texture2D.MostDetailedMip = 0;
                    g_pd3dDevice->CreateShaderResourceView(texture, &srvDesc, &image);
                }
                if (texture && image)
                {
                    BYTE* pData = NULL;

                    D3D11_MAPPED_SUBRESOURCE mappedResource{};
                    HRESULT hr = g_pImmediateContext->Map(texture, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
                    if (SUCCEEDED(hr))
                    {
                        WIC->CopyPixels(
                            NULL,
                            duplicator->Width * 4,
                            duplicator->Width * duplicator->Height * 4,
                            (BYTE*)mappedResource.pData);
                        g_pImmediateContext->Unmap(texture, 0);
                    }
                    if (image)
                    {
                        auto pos = ImGui::GetCursorPos();
                        float ww = windowSize.cx - (pos.x + 5);
                        float hh = ww * 0.5625;
                        ImGui::Image(image, ImVec2(ww, hh), ImVec2(0.0f, 1.0f), ImVec2(1.0f, 0.0f));
                        double SoundTime = (double(now) - double(StartTime)) / 1000.0;
                        int hours = SoundTime / 3600;
                        int remainingSeconds = fmod(SoundTime, 3600);
                        int minutes = remainingSeconds / 60;
                        int seconds = remainingSeconds % 60;
                        int mseconds = fmod(SoundTime, 1) * 1000;
                        auto tx = StringHelper::Format("%02d:%02d:%02d,%03d",
                            hours, minutes, seconds, mseconds
                        );
                        ImGui::Text(tx.c_str());
                    }
                    else
                    {
                        ImGui::Text(u8s("未开始"));
                    }
                }
            }
        }
    }
    else
    {
        auto pos = ImGui::GetCursorPos();
        float ww = windowSize.cx - (pos.x + 5);
        float hh = ww * 0.5625;
        ImDrawRect(pos.x, pos.y, ww, hh, ImColor(1.0f, 1.0f, 1.0f, 1.0f), 2.5f);
        ImGui::SetCursorPos(ImVec2(pos.x, pos.y + hh + 5));
        ImGui::Text(u8s("未开始"));
    }

}
void Render()
{
    g_pImmediateContext->OMSetRenderTargets(1, &g_pRenderTargetView, nullptr);
    float black[] = { 0.0f, 0.0f, 0.0f, 0.0f };
    g_pImmediateContext->ClearRenderTargetView(g_pRenderTargetView, black);
    int width, height;
    RECT rect;
    GetClientRect(hwnd, &rect);
    width = rect.right - rect.left;
    height = rect.bottom - rect.top;

    ImGui_ImplDX11_NewFrame();
    ImGui_ImplWin32_NewFrame();
    ImGui::NewFrame();
    ImGui::SetNextWindowSize(ImVec2((float)width, (float)height));
    ImGui::SetNextWindowPos(ImVec2(0, 0));
    ImGui::Begin(u8s("##主窗口"), nullptr,
        ImGuiWindowFlags_NoTitleBar |
        ImGuiWindowFlags_NoResize |
        ImGuiWindowFlags_NoMove |
        ImGuiWindowFlags_NoScrollbar |
        ImGuiWindowFlags_NoScrollWithMouse);
    {
        RenderGui({ width, height });
    }
    ImGui::End();
    ImGui::EndFrame();
    ImGui::Render();
    ImGui_ImplDX11_RenderDrawData(ImGui::GetDrawData());
    g_pSwapChain->Present(0, 0);
}
void InitD3D(HWND hWnd) {
    DXGI_SWAP_CHAIN_DESC sd;
    ZeroMemory(&sd, sizeof(sd));
    sd.BufferCount = 1;
    sd.BufferDesc.Format = DXGI_FORMAT_B8G8R8A8_UNORM;
    sd.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
    sd.OutputWindow = hWnd;
    sd.SampleDesc.Count = 1;
    sd.Windowed = TRUE;

    D3D_FEATURE_LEVEL featureLevel;
    const D3D_FEATURE_LEVEL featureLevels[] = { D3D_FEATURE_LEVEL_11_0 };

    D3D11CreateDeviceAndSwapChain(nullptr, D3D_DRIVER_TYPE_HARDWARE, nullptr, 0, featureLevels, 1,
        D3D11_SDK_VERSION, &sd, &g_pSwapChain, &g_pd3dDevice, &featureLevel, &g_pImmediateContext);
}
void CreateRenderTarget() {
    ID3D11Texture2D* pBackBuffer = nullptr;
    g_pSwapChain->GetBuffer(0, IID_PPV_ARGS(&pBackBuffer));

    D3D11_RENDER_TARGET_VIEW_DESC desc;
    ZeroMemory(&desc, sizeof(desc));
    desc.Format = DXGI_FORMAT_B8G8R8A8_UNORM;
    desc.ViewDimension = D3D11_RTV_DIMENSION_TEXTURE2D;

    g_pd3dDevice->CreateRenderTargetView(pBackBuffer, &desc, &g_pRenderTargetView);
    pBackBuffer->Release();
}

extern LRESULT ImGui_ImplWin32_WndProcHandler(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
    ImGui_ImplWin32_WndProcHandler(hWnd, message, wParam, lParam);
    switch (message) {
    case WM_DESTROY:
        exit(0);
        break;
    case WM_SIZE:
        if (g_pSwapChain)
        {
            ImGui::GetIO().DisplaySize = ImVec2((float)LOWORD(lParam), (float)HIWORD(lParam));/*
            if (g_pRenderTargetView) {
                g_pRenderTargetView->Release();
                g_pRenderTargetView = nullptr;
            }*/
            ID3D11Texture2D* pBackBuffer = nullptr;
            HRESULT hr = g_pSwapChain->ResizeBuffers(0, 0, 0, DXGI_FORMAT_UNKNOWN, 0);
            if (SUCCEEDED(hr)) {
                hr = g_pSwapChain->GetBuffer(0, IID_PPV_ARGS(&pBackBuffer));
                if (SUCCEEDED(hr)) {
                    g_pd3dDevice->CreateRenderTargetView(pBackBuffer, nullptr, &g_pRenderTargetView);
                    pBackBuffer->Release();
                }
            }
            Render();
        }
        break;
    case WM_MOVE:
        if (g_pSwapChain)
        {
            Render();
        }
        break;
    default:
        return DefWindowProc(hWnd, message, wParam, lParam);
    }
    return 0;
}
void ConfigWindow()
{
    TCHAR exePath[MAX_PATH]{};
    GetModuleFileName(GetModuleHandle(NULL), exePath, MAX_PATH);
    HICON hIcon = ExtractIcon(GetModuleHandle(NULL), exePath, 0);
    WNDCLASSEX wc = {};
    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.hInstance = GetModuleHandleA(NULL);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOWFRAME);
    wc.hIcon = hIcon;
    wc.hIconSm = hIcon;
    wc.lpszClassName = L"KNWindowClass";
    RegisterClassEx(&wc);
    int width = 300;
    int height = 360;
    hwnd = CreateWindow(
        L"KNWindowClass",
        L"码德录屏",
        WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU,
        CW_USEDEFAULT, CW_USEDEFAULT,
        width, height,
        NULL, NULL,
        GetModuleHandle(NULL),
        NULL);
    SetClassLongPtr(hwnd, GCLP_HICON, (LONG_PTR)LoadIcon(NULL, IDI_APPLICATION));
}
void InitImGui() {
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO();
    (void)io;
    ImGui_ImplWin32_Init(hwnd);
    ImGui_ImplDX11_Init(g_pd3dDevice, g_pImmediateContext);
    ImGui_ImplDX11_CreateDeviceObjects();
    ImGui::GetIO().ImeWindowHandle = hwnd;
}
void CleanupImGui() {
    ImGui_ImplDX11_Shutdown();
    ImGui_ImplWin32_Shutdown();
    ImGui::DestroyContext();
}
bool ConfigDXGIDuplicator()
{
    HRESULT hr = D3D11CreateDevice(nullptr, D3D_DRIVER_TYPE_HARDWARE, nullptr, 0, nullptr, 0, D3D11_SDK_VERSION, &pD3DDevice_Duplicator, nullptr, &pContext_Duplicator);
    if (FAILED(hr))
        return false;
    //硬件加速桌面捕获对象
    duplicator = new DXGIDuplicator();
    if (!duplicator->InitD3D11Device(pD3DDevice_Duplicator, pContext_Duplicator))
        return false;
    if (!duplicator->InitDuplication())
        return false;
    return true;
}
void ConfigCursorData()
{
    GENERATE_CURSOR_DATA(32512);
    GENERATE_CURSOR_DATA(32513);
    GENERATE_CURSOR_DATA(32514);
    GENERATE_CURSOR_DATA(32515);
    GENERATE_CURSOR_DATA(32516);
    GENERATE_CURSOR_DATA(32642);
    GENERATE_CURSOR_DATA(32643);
    GENERATE_CURSOR_DATA(32644);
    GENERATE_CURSOR_DATA(32645);
    GENERATE_CURSOR_DATA(32646);
    GENERATE_CURSOR_DATA(32648);
    GENERATE_CURSOR_DATA(32649);
    GENERATE_CURSOR_DATA(32650);
    GENERATE_CURSOR_DATA(32651);
    GENERATE_CURSOR_DATA(32671);
    GENERATE_CURSOR_DATA(32672);
}
int main(int argc, const char** arg)
{
    ConfigWindow();
    ConfigCursorData();
    InitD3D(hwnd);
    CreateRenderTarget();
    InitImGui();
    ConfigDXGIDuplicator();
    ShowWindow(hwnd, true);
    MSG msg = {};
    bool running = true;
    while (running)
    {
        if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        {
            if (msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
            Render();
        }
        static LONGLONG lastRender = 0;
        LONGLONG now = GetTickCount64();
        if (now - lastRender > 10)
        {
            lastRender = now;
            Render();
            continue;
        }
        Sleep(1);
    }
    g_pRenderTargetView->Release();
    g_pSwapChain->Release();
    g_pImmediateContext->Release();
    g_pd3dDevice->Release();
    return NULL;
}
